
import vrml.*;
import vrml.field.*;
import vrml.node.*;

class Triangle
{
	public float v1[];
	public float v2[];
	public float v3[];
	
	public Triangle(float v1[], float v2[], float v3[])
	{
		this.v1 = v1;
		this.v2 = v2;
		this.v3 = v3;
	}
	
	public void print()
	{
		System.out.printf("%f %f %f   %f %f %f   %f %f %f",
			v1[0], v1[1], v1[2],
			v2[0], v2[1], v2[2],
			v3[0], v3[1], v3[2]
		);
	}
}

class Math
{
	public static float[] subtract(float lhs[], float rhs[])
	{
		float ret[] = new float[3];
		for(int i = 0; i < 3; i++)
			ret[i] = lhs[i] - rhs[i];
		return ret;
	}
	
	public static float dot(float lhs[], float rhs[])	
	{
		float ret = 0.0f;
		for(int i = 0; i < 3; i++)
			ret += lhs[i]*rhs[i];
		return ret;
	}
	
	public static float[] cross(float lhs[], float rhs[])
	{
		float ret[] = new float[3];
		ret[0] = lhs[1]*rhs[2]-lhs[2]*rhs[1];
		ret[1] = lhs[2]*rhs[0]-lhs[0]*rhs[2];
		ret[2] = lhs[0]*rhs[1]-lhs[1]*rhs[0];
		return ret;
	}
	
	public static boolean testRayTriangle(float orig[], float dir[], float v0[], float v1[], float v2[])
	{
		// nach Möller
		float edge1[] = subtract(v1, v0);
		float edge2[] = subtract(v2, v0);
		float pvec[] = cross(dir, edge2);
		float det = dot(edge1, pvec);
		
		float eps = 0.000001f;
		if(det > -eps && det < eps)
			return false;
		
		float inv_det = 1.0f/det;
		float tvec[] = subtract(orig, v0);
		float u = dot(tvec, pvec)*inv_det;
		if(u < 0.0f || u > 1.0f)
			return false;
		
		float qvec[] = cross(tvec, edge1);
		float v = dot(dir, qvec)*inv_det;
		if(v < 0.0f || (u+v) > 1.0f)
			return false;
		
		float t = dot(edge2, qvec)*inv_det;
		
		if(t > 0.0f)
			return true;
		
		return false;
	}
	
	public static float[] translate(float p[], float t[])
	{
		float ret[] = new float[3];
		for(int i = 0; i < 3; i++)
			ret[i] = p[i] + t[i];
		return ret;
	}

	public static Triangle translate(Triangle t, float trans[])
	{
		return new Triangle(
			translate(t.v1, trans),
			translate(t.v2, trans),
			translate(t.v3, trans)
		);
	}
	
	public static float[] rotate(float p[], float axis[], float angle)
	{
		// TODO: implement rotation for vertex
		return null;
	}
	
	public static float[] rotate(Triangle t, float axis[], float angle)
	{
		// TODO: implement rotation for triangle
		return null;
	}
}

public class Collision extends Script
{
	private boolean collCheck = false;
	
	private SFTime soundStartTime;
	private ConstSFTime cycleTime;
	
	private MFVec3f tuch_Points_N;
	private MFInt32 carriage1_Index_N;
	private MFVec3f carriage1_Points_N;
	private MFInt32 carriage2_Index_N;
	private MFVec3f carriage2_Points_N;
	
	private SFVec3f tuch_Translation;
	//private SFVec3f tuch_Scale;
	//private SFRotation tuch_Rotation;

	private SFVec3f carr_Translation;
	//private SFVec3f carr_Scale;
	//private SFRotation carr_Rotation;
	
	private float tuch_Points[][];
	private Triangle carriage_Triangles[];
	
	
	private void check()
	{
            System.out.println("JAVA: check");
            // hier muss die kollisionserkennung hin
 
//             // funktion fuer test Strahl Dreieck ist bereits implementiert :)
//             // Bsp.: in PSEUDO-CODE!!!!
//             vertex1 = new SFVec3f(carriageGeo1.children[0].geometry.coord.point[0][0], 
//                 carriageGeo1.children[0].geometry.coord.point[0][1], 
//                 carriageGeo1.children[0].geometry.coord.point[0][2]);
//             vertex2 = new SFVec3f(carriageGeo1.children[0].geometry.coord.point[1][0], 
//                 carriageGeo1.children[0].geometry.coord.point[1][1], 
//                 carriageGeo1.children[0].geometry.coord.point[1][2]);
//             vertex3 = new SFVec3f(carriageGeo1.children[0].geometry.coord.point[2][0], 
//                 carriageGeo1.children[0].geometry.coord.point[2][1], 
//                 carriageGeo1.children[0].geometry.coord.point[2][2]);
//  
//             testPoint = new SFVec3f(ragGeo.children[0].geometry.coord.point[0][0], 
//                 ragGeo.children[0].geometry.coord.point[0][1], 
//                 ragGeo.children[0].geometry.coord.point[0][2]);
//  
//             float dir[] = {1.0f, 0.0f, 0.0f};
//             // TEST ist bool (TRUE / FALSE)
//             TEST = testPointTriange(testPoint, dir, vertex1, vertex2, vertex3);
//  
//             if (TEST == TRUE) 
//             {
//                 sound startTime
//             } else 
//             {
//                 print('keine kollision')
//             }   

	
	}
	
	@Override
	public void initialize()
	{
		System.out.println("JAVA: init");
		
		SFNode carriageGeo1 = (SFNode)getField("carriageGeo1");
		SFNode carriageGeo2 = (SFNode)getField("carriageGeo2");
		SFNode ragGeo = (SFNode)getField("ragGeo");
		SFNode carriageTrans = (SFNode)getField("carriageTrans");
		SFNode ragTrans = (SFNode)getField("ragTrans");

		// get time fields
		SFNode sound = (SFNode)getField("sound");
		SFNode time = (SFNode)getField("time");
		soundStartTime = (SFTime)((Node)sound.getValue()).getEventIn("startTime");
		cycleTime = (ConstSFTime)((Node)time.getValue()).getEventOut("cycleTime");

		// get rag point field
		MFNode n1 = (MFNode)((Node)ragGeo.getValue()).getExposedField("children");
		SFNode n2 = (SFNode)((Node)n1.get1Value(0)).getExposedField("geometry");
		SFNode n3 = (SFNode)((Node)n2.getValue()).getExposedField("coord");
		tuch_Points_N = (MFVec3f)((Node)n3.getValue()).getExposedField("point");
		
		// get carriage data
		n1 = (MFNode)((Node)carriageGeo1.getValue()).getExposedField("children");
		n2 = (SFNode)((Node)n1.get1Value(0)).getExposedField("geometry");
		n3 = (SFNode)((Node)n2.getValue()).getExposedField("coord");
		carriage1_Points_N = (MFVec3f)((Node)n3.getValue()).getExposedField("point");
		
		n1 = (MFNode)((Node)carriageGeo1.getValue()).getExposedField("children");
		n2 = (SFNode)((Node)n1.get1Value(0)).getExposedField("geometry");
		carriage1_Index_N = (MFInt32)((Node)n2.getValue()).getExposedField("coordIndex");

		n1 = (MFNode)((Node)carriageGeo2.getValue()).getExposedField("children");
		n2 = (SFNode)((Node)n1.get1Value(0)).getExposedField("geometry");
		n3 = (SFNode)((Node)n2.getValue()).getExposedField("coord");
		carriage2_Points_N = (MFVec3f)((Node)n3.getValue()).getExposedField("point");
		
		n1 = (MFNode)((Node)carriageGeo2.getValue()).getExposedField("children");
		n2 = (SFNode)((Node)n1.get1Value(0)).getExposedField("geometry");
		carriage2_Index_N = (MFInt32)((Node)n2.getValue()).getExposedField("coordIndex");
		
		// get "tuch" transform
		tuch_Translation = (SFVec3f)((Node)ragTrans.getValue()).getExposedField("translation");
		//tuch_Scale = (SFVec3f)((Node)ragTrans.getValue()).getExposedField("scale");
		//tuch_Rotation = (SFRotation)((Node)ragTrans.getValue()).getExposedField("rotation");
		
		// get carriage transform
		carr_Translation = (SFVec3f)((Node)carriageTrans.getValue()).getExposedField("translation");
		//carr_Scale = (SFVec3f)((Node)carriageTrans.getValue()).getExposedField("scale");
		//carr_Rotation = (SFRotation)((Node)carriageTrans.getValue()).getExposedField("rotation");
		
		// copy "tuch" vertex data
		tuch_Points = new float[tuch_Points_N.getSize()][];
		for(int i = 0; i < tuch_Points.length; i++)
			tuch_Points[i] = new float[3];
		tuch_Points_N.getValue(tuch_Points);
		
		float trans[] = new float[3];
		tuch_Translation.getValue(trans);
		for(int i = 0; i < tuch_Points.length; i++)
			tuch_Points[i] = Math.translate(tuch_Points[i], trans);
		
		// TODO: apply "tuch" scale
		
		// copy carriage vertex data
		int idxs[] = new int[carriage1_Index_N.getSize()];
		carriage1_Index_N.getValue(idxs);
		int idxs2[] = new int[carriage2_Index_N.getSize()];
		carriage2_Index_N.getValue(idxs2);

		int nTriag = idxs.length / 4;
		int nTriag2 = idxs2.length / 4;
		carriage_Triangles = new Triangle[nTriag+nTriag2];
		
		for(int i = 0; i < idxs.length; i+=4)
		{
			float v1[] = new float[3];
			float v2[] = new float[3];
			float v3[] = new float[3];
			
			if(idxs[i+3] != -1)
			{
				System.out.println("error!\n");
				return;
			}
			
			carriage1_Points_N.get1Value(idxs[i], v1);
			carriage1_Points_N.get1Value(idxs[i+1], v2);
			carriage1_Points_N.get1Value(idxs[i+2], v3);
			
			carriage_Triangles[i/4] = new Triangle(v1, v2, v3);
		}

		for(int i = 0; i < idxs2.length; i+=4)
		{
			float v1[] = new float[3];
			float v2[] = new float[3];
			float v3[] = new float[3];
			
			if(idxs2[i+3] != -1)
			{
				System.out.println("error!\n");
				return;
			}
			
			carriage2_Points_N.get1Value(idxs2[i], v1);
			carriage2_Points_N.get1Value(idxs2[i+1], v2);
			carriage2_Points_N.get1Value(idxs2[i+2], v3);
			
			carriage_Triangles[nTriag+i/4] = new Triangle(v1, v2, v3);
		}

		
	}
	
	@Override
	public void processEvent(Event e)
	{
		String name = e.getName();
		if(name.equals("cycle"))
		{
			System.out.println("JAVA: cycle event called");
			if(collCheck)
				check();
		}
		else if(name.equals("actionPress"))
		{
			ConstSFString s = (ConstSFString)e.getValue();
			if(s.getValue().equals("<"))
			{
				System.out.println("JAVA: key '<' pressed!");
				collCheck = !collCheck;
			}
		}
	}

}



